<?php

declare(strict_types=1);

namespace Drupal\devel_wizard\Plugin\DevelWizard\Spell;

use Drupal\block_content\BlockContentTypeInterface;
use Drupal\block_content\Entity\BlockContentType;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Extension\ModuleExtensionList;
use Drupal\Core\Extension\ModuleInstallerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\Core\StringTranslation\TranslationInterface;
use Drupal\devel_wizard\Attribute\DevelWizardSpell;
use Drupal\devel_wizard\ShellProcessFactoryInterface;
use Drupal\devel_wizard\Spell\SpellTraitPackageManager;
use Drupal\devel_wizard\Utils;
use Psr\Log\LoggerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * @todo Common parent class for:
 *   - BlockContentTypeCreateSpell
 *   - NodeTypeCreateSpell
 *   - TaxonomyVocabularyCreateSpell
 */
#[DevelWizardSpell(
  id: 'devel_wizard_block_content_type_create',
  category: new TranslatableMarkup('Block type'),
  label: new TranslatableMarkup('Block type - Create'),
  description: new TranslatableMarkup('Creates a new Block type without any extra.'),
  tags: [
    'bundle' => new TranslatableMarkup('Bundle'),
    'config_entity' => new TranslatableMarkup('Config entity'),
    'block_content_type' => new TranslatableMarkup('Block type'),
    'site_build' => new TranslatableMarkup('Site build'),
  ],
)]
class BlockContentTypeCreateSpell extends ConfigEntitySpellBase {

  use SpellTraitPackageManager;

  /**
   * {@inheritdoc}
   */
  protected string $provider = 'block_content';

  protected string $configEntityTypeId = 'block_content_type';

  protected string $contentEntityTypeId = 'block_content';

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('messenger'),
      $container->get('logger.channel.devel_wizard_spell'),
      $container->get('string_translation'),
      $container->get('devel_wizard.utils'),
      $container->get('config.factory'),
      $container->get('entity_type.manager'),
      $container->get('devel_wizard.shell_process_factory'),
      $container->get('module_installer'),
      $container->get('extension.list.module'),
    );
  }

  public function __construct(
    array $configuration,
    $plugin_id,
    $plugin_definition,
    MessengerInterface $messenger,
    LoggerInterface $logger,
    TranslationInterface $stringTranslation,
    Utils $utils,
    ConfigFactoryInterface $configFactory,
    EntityTypeManagerInterface $entityTypeManager,
    ShellProcessFactoryInterface $shellProcessFactory,
    ModuleInstallerInterface $moduleInstaller,
    ModuleExtensionList $moduleList,
  ) {
    $this->entityTypeManager = $entityTypeManager;
    $this->shellProcessFactory = $shellProcessFactory;
    $this->moduleInstaller = $moduleInstaller;
    $this->moduleList = $moduleList;
    parent::__construct(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $messenger,
      $logger,
      $stringTranslation,
      $utils,
      $configFactory,
    );
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [
      'values' => [
        'id' => '',
        'label' => '',
        'description' => 'Useless description',
        'help' => 'Useless help',
        'revision' => 0,
      ],
    ];
  }

  protected function populateCalculatedConfigurationValues(): static {
    parent::populateCalculatedConfigurationValues();
    if (empty($this->configuration['values']['id'])) {
      throw new \InvalidArgumentException('id is required');
    }

    if (empty($this->configuration['values']['label'])) {
      $this->configuration['values']['label'] = $this->configuration['values']['id'];
    }

    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $configuration = $this->getConfiguration();

    $form['values'] = [
      '#tree' => TRUE,
      'id' => [
        '#type' => 'machine_name',
        '#required' => TRUE,
        '#default_value' => $configuration['values']['id'],
        '#maxlength' => 64,
        '#description' => $this->t('A unique machine-name for this Block type. It must only contain lowercase letters, numbers, and underscores.'),
        '#machine_name' => [
          'exists' => [static::class, 'blockContentTypeLoad'],
          'standalone' => TRUE,
        ],
        '#autocomplete_route_name' => 'devel_wizard.autocomplete.config_entity_instance',
        '#autocomplete_route_parameters' => [
          'entityTypeId' => $this->configEntityTypeId,
        ],
      ],
    ];

    return $form;
  }

  public static function blockContentTypeLoad(string $id): ?BlockContentTypeInterface {
    return class_exists(BlockContentType::class) ?
      BlockContentType::load($id)
      : NULL;
  }

  /**
   * {@inheritdoc}
   */
  public function validateConfigurationForm(array &$form, FormStateInterface $form_state) {
    // @todo Implement validateConfigurationForm() method.
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    $values = $form_state->getValue($form['#parents'], []);
    $this->setConfiguration($values);
  }

  /**
   * @throws \Drupal\Core\Entity\EntityStorageException
   * @throws \Drupal\Core\Entity\EntityMalformedException
   */
  protected function doIt(): static {
    if (!isset($this->batchContext['sandbox']['current_step'])) {
      $this->batchContext['sandbox']['current_step'] = 'required_modules_install';
    }

    $configuration = $this->getConfiguration();
    switch ($this->batchContext['sandbox']['current_step']) {
      case 'required_modules_install':
        $requiredModules = array_keys($this->getRequiredModules(), TRUE);
        if ($requiredModules) {
          foreach ($requiredModules as $moduleName) {
            $this->ensureModuleInstalled($moduleName);
          }
          // @todo Clear the cache only when it is necessary.
          drupal_flush_all_caches();
        }

        $this->batchContext['message'] = $this->t(
          'Create Block type: @entity.label',
          [
            '@entity.label' => $configuration['values']['label'],
          ],
        );
        $this->batchContext['sandbox']['current_step'] = 'config_entity_instance_create';
        $this->batchContext['sandbox']['finished'] = 0.5;
        break;

      case 'config_entity_instance_create':
        $this->createConfigInstance($this->configuration['values']);
        $this->batchContext['sandbox']['current_step'] = '_finished';
        $this->batchContext['sandbox']['finished'] = 1.0;
        break;

      case '_finished':
        $this->batchContext['sandbox']['finished'] = 1.0;
        break;

    }

    return $this;
  }

}
